package asp;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 * In the current version this implementation cannot make use of all features
 * theoretically provided by oClingo because an important one is buggy and does
 * not work. So this ASP classes implement some workarounds at different places,
 * to use the oClingo application anyway.
 * 
 * @author Alexander Häusler
 * @author Marco Müller
 * 
 * @version 1.0
 */
public class XASPController {
	
	private String matchID;
	private String role;
	
	protected ArrayList<Step> steps;
	protected ArrayList<Step> externalSteps;
	protected Iterator<Step>  stepsIterator;
	
	private int stepCounter = 0;
	
	private XASPSolverController solver;
	
	private static final String ROLE_PATTERN = "\\(\\s*role\\s*\\w+\\s*\\)";
	
	
	

	public XASPController(String matchID, String role) {
		this.matchID  = matchID;
		this.role	  = role;
		steps 		  = new ArrayList<Step>();
		externalSteps = new ArrayList<Step>();
		solver 		  = new XASPSolverController(this, role, "./asp/" + this.matchID + ".lp");
		stepsIterator = steps.iterator();
	}
	
	/**
	 * reads the given GDL and proves if it is a singleplayer game
	 * 
	 * @return
	 * @throws ASPException
	 */
	public boolean isSoloGame() throws ASPException {
		
		String buffer;
		
		try {
			String filePath = "./kif/" + matchID + ".kif";
			FileInputStream inputStream = new FileInputStream(filePath);
			InputStreamReader inputStreamReader = new InputStreamReader(inputStream);
			BufferedReader br = new BufferedReader(inputStreamReader);
			int rolecount = 0;
			
			while ((buffer = br.readLine()) != null) {
				Pattern pattern = Pattern.compile(ROLE_PATTERN, Pattern.CASE_INSENSITIVE);
				Matcher m = pattern.matcher(buffer);
				
				while (m.find())
					rolecount++;
			}
			
			br.close();
			
			return (rolecount == 1);
			

		} catch(IOException e) {
			throw new ASPException("roles could not been identified!");
		}
	}
	
	
	/**
	 * Classes using this XASPController can - after the game is solved -
	 * use this method to get the solution step by step.
	 * 
	 * @return The next step as a GDL string
	 * @throws ASPException When the given game is not solved yet
	 */
	public String getNextStep() throws ASPException {
		if (!this.isSolved())
			throw new ASPException("steps are not ready!");
		
		for (int i = 0; i < this.stepCounter; i++)
			stepsIterator.next();
		
		if (stepsIterator.hasNext()) {
			Step tmp = stepsIterator.next();
			
			System.out.println(tmp.toASPString());
			String str = tmp.toGDLString();
			return str;
		}
		else
			return null;
	}
	
	
	public void putInStep(String step) {
		solver.putInStep(Step.parseGDLString(step));
		this.stepCounter++;
	}
	
	
	public void kill() {
		solver.kill();
	}
	
	
	public void restart() {
		solver.kill();
		
		// new Thread
		solver = new XASPSolverController(this, this.role, "./asp/" + this.matchID + ".lp");
		solver.start();
	}
	
	
	public boolean isSolved() {
		return this.solver.isSolved();
	}
	
	/**
	 * Translates GDL to oclingo encoding by using ASPTranslator and starts solving process
	 */
	public void start() {
		// prepare gdl parser and set it to oclingo mode
		XASPEncodingConstructor parser = new XASPEncodingConstructor("oclingo");
		
		try {
			
			FileReader  reader = new FileReader (new File("./kif/" + matchID + ".kif"));
			PrintStream writer = new PrintStream(new File("./asp/" + matchID + ".lp"));

			
			// redirect stdout to /dev/null
			PrintStream stdoutBackup = System.out;
			System.setOut(new PrintStream("/dev/null"));
			
			// parse gdl to oClingo
			parser.translate(reader, writer);
			
			// restore stdout
			System.setOut(stdoutBackup);
			
			
		
			// the solving process is started
			solver.run();
			
		} catch(Exception e) {
			e.printStackTrace();
			System.err.println("XASPController could not be started!");
		}
		
	}
	

}
