package basic.ai;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import basic.logic.Move;
import basic.logic.State;
import basic.prepro.STTranslator;

public class Node {

	private final State state;

	private HashMap<Move, Node> childNodes = new HashMap<Move, Node>();
	private Node parent;

	private boolean isTerminal;

	private int tnodes;

	private boolean _cache_fullyExpanded = false;

	private List<Move> legalMoves;

	private int visits = 0;
	private int total_visited = 0;
	private int rewardSum = 0;

	public Node(State state, List<Move> legalMoves, Node parentNode,
			boolean isTerminal) {
		this.state = state;

		this.parent = parentNode;
		this.legalMoves = legalMoves;

		this.isTerminal = isTerminal;

		this.tnodes = 1;
	}

	public void setParent(Node node) {
		this.parent = parent;
	}

	public Node getParent() {
		return parent;
	}

	public Node getChild(Move move) {
		return childNodes.get(move);
	}

	public Node addChild(Move move, State nodesState, List<Move> legalMoves,
			boolean isTerminal) {
		childNodes
				.put(move, new Node(nodesState, legalMoves, this, isTerminal));
		++tnodes;
		return childNodes.get(move);

	}

	public Collection<Node> getChildren() {
		return childNodes.values();
	}

	public Collection<Move> getExtendedMoves() {
		return childNodes.keySet();
	}

	public Collection<Move> getUnextendedMoves() {
		Collection<Move> returnCollection = legalMoves;
		returnCollection.removeAll(childNodes.keySet());
		return returnCollection;
	}

	public boolean isTerminal() {
		return isTerminal;
	}

	public boolean isFullyExpanded() {
		return _cache_fullyExpanded || checkFullyExpanded();
	}

	private boolean checkFullyExpanded() {
		if (childNodes.keySet().containsAll(legalMoves)) {
			_cache_fullyExpanded = true;
		}

		return _cache_fullyExpanded;
	}

	public int getVisits() {
		return visits;
	}

	public int getRewardSum() {
		return rewardSum;
	}

	public void addReward(int reward) {
		visits++;
		rewardSum += reward;
	}

	public State getState() {
		return state;
	}

	public int total_nodes(){
		return tnodes;
	}

	public int total_visitedn(){
		return total_visited;
	}
}