package basic.ai;

import basic.logic.GameLogic;
import basic.logic.Move;
import basic.logic.Role;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

/**
 * basic General Game Player from Potsdam
 *
 * Copyright (C) 2011 Marius Schneider
 *
 * The Player is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * The Player is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with Centurio. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This class implements a simple Monte Carlo Strategy
 * 
 */
public class SimpleMC extends Thread implements AI {

	static basic.logic.State currentState = null;
	static Boolean newState = true;

	// example
	int[] visits = null; /* number of visits per move */
	int[] means = null; /* sum of all random games per move */
	List<Move> moves = null; /* list of all moves */
	int numberOfMoves = 0;
	Role ourRole = null;

	/** Game Logic to interacte with game description */
	public GameLogic gl = null;

	/**
	 * Constructor
	 */
	public SimpleMC(GameLogic gl, Role ourRole) {
		this.gl = gl;
		this.ourRole = ourRole;
	}

	/**
	 * standard run-method of a thread <br>
	 * infinite loop, which can only terminate with a interrupt <br>
	 */
	public void run() {

		basic.logic.State stateTemp = null;
		Random rand = new Random();
		int playerIndex = gl.findIndex(ourRole);
		// while time is left
		while (!isInterrupted()) {
			// global state of game changed
			if (newState) {
				System.out.println("State was updated!");
				stateTemp = currentState;
				newState = false;
				// get leag moves
				moves = gl.legal(currentState);
				numberOfMoves = moves.size();
				// initial stat arrays
				visits = new int[numberOfMoves];
				means = new int[numberOfMoves];
			}
			// for each move
			for (int index = 0; index < moves.size(); index++) {
				// play a random game and sum up all reached scores
				means[index] += randomGame(stateTemp, moves.get(index), rand)[playerIndex];
				visits[index]++;
			}

		}

		this.gl.destroy();

	}

	/**
	 * plays a random game
	 * 
	 * @param currentNode
	 *            current State
	 * @param nextMove
	 *            List of possible moves
	 * @param rand
	 *            random generator
	 * @return returns score of first role
	 */
	private int[] randomGame(basic.logic.State currentNode, Move nextMove,
			Random rand) {
		basic.logic.State nextNode = gl.nextState(currentNode, nextMove);
		while (!gl.isTerminal(nextNode)) {
			// all legal moves
			List<Move> moves = gl.legal(nextNode);
			// randomly choose one move
			Move rMove = moves.get(Math.abs(rand.nextInt()) % moves.size());
			// calculate next state
			nextNode = gl.nextState(nextNode, rMove);
		}
		return gl.goal(nextNode);
	}

	/**
	 * set new state from the new move of the Game Message
	 * 
	 * @param nextState
	 */
	public void setNewState(basic.logic.Move executedMove) {
		currentState = gl.nextState(currentState, executedMove);
		newState = true;
	}

	public void initState(basic.logic.State initState) {
		currentState = initState;
		newState = true;
	}

	/**
	 * evaluate the best move <br>
	 * important for GamePlayer Play Message
	 * 
	 * @param role
	 *            Role of basicplayer
	 * @return best move
	 */
	public Object getBestMove(Role role) {

		double bestmean = -1.0;
		Move bestMove = null;
		// for each move
		for (int index = 0; index < numberOfMoves; index++) {
			String move = basic.prepro.STTranslator.printKIF(moves.get(index)
					.get());
			System.out.println("Move: " + move + " Score: " + means[index]
					/ visits[index] + " Visits: " + visits[index]);
			// remember move with highest averaged score
			if (means[index] / visits[index] > bestmean) {
				bestmean = means[index] / visits[index];
				bestMove = moves.get(index);
			}

		}
	LinkedList<Object> oMoves = (LinkedList<Object>) bestMove.get();
	System.out.println("Chosen: " + basic.prepro.STTranslator.printKIF(oMoves.get(0)));
	int moveID = gl.findIndex(role);
		return oMoves.get(moveID);
		//return bestMove.get();
	}

}
