package basic.ai.mcts;

import basic.ai.AI;
import basic.ai.Node;
import basic.logic.GameLogic;
import basic.logic.Move;
import basic.logic.PrologLogic;
import basic.logic.Role;
import basic.logic.State;

import java.util.LinkedList;
import java.util.List;
import java.util.Random;

/**
 * basic General Game Player from Potsdam
 *
 * Copyright (C) 2011 Marius Schneider
 *
 * The Player is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * The Player is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with Centurio. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This class implements a Monte Carlo Tree Search Strategy
 * 
 */

public class MCTS extends Thread implements AI {

	static Boolean newState = true;

	private Node root;
	Role ourRole = null;

	/** Game Logic to interact with game description */
	public GameLogic gl = null;

	// policies needed for mcts
	private TreePolicy treePolicy;
	private DefaultPolicy defaultPolicy;
	private Backpropagation backpropagation;
	
	/**
	 * Constructor
	 */
	public MCTS(GameLogic gl, Role ourRole) {
		this.gl = gl;
		this.ourRole = ourRole;
		
		treePolicy = new UCTSelectionPolicy(gl);
		defaultPolicy = new RandomPlayoutPolicy(gl);
		backpropagation = new DefaultBackpropagation();
	}

	/**
	 * standard run-method of a thread <br>
	 * infinite loop, which can only terminate with a interrupt <br>
	 */
	public void run(){

		int playerIndex = gl.findIndex(ourRole);
		// while time is left
		while (!isInterrupted()) {
			Node nextNode = treePolicy.selectNode(root);
			int[] reward = defaultPolicy.playout(nextNode);
			backpropagation.backup(nextNode, reward[playerIndex]);
		}
		this.gl.destroy();
	}

	/**
	 * 
	 */
	public void initState(basic.logic.State initState){
		root = new Node(initState, gl.legal(initState), null, gl.isTerminal(initState));
		newState = true;
	}

	/**
	 * set new state from the new move of the Game Message
	 * 
	 * @param executedMove
	 */
	public void setNewState(basic.logic.Move executedMove) {
		root = root.getChild(executedMove);
		root.setParent(null); //remove parent reference, since this is our new root;
		newState = true;
	}

	/**
	 * evaluate the best move <br>
	 * important for GamePlayer Play Message
	 * 
	 * @param role
	 *            Role of basicplayer
	 * @return best move
	 */
	public Object getBestMove(Role role) {

		int maxVisits = 0;
		Move bestMove = null;
		
		for (Move curMove : root.getExtendedMoves()) {
			int visits = root.getChild(curMove).getVisits();
			if (visits > maxVisits) {
				maxVisits = visits;
				bestMove = curMove;
			}
		}
		LinkedList<Object> linkedList = (LinkedList<Object>) bestMove.get();
		LinkedList<Object> oMoves = linkedList;
		System.out.println("Chosen: " + basic.prepro.STTranslator.printKIF(oMoves.get(0)));
		int moveID = gl.findIndex(role);
		return oMoves.get(moveID);
//		return bestMove.get();
	}

}
