package basic.ai.mcts;

import java.util.List;
import java.util.Random;

import basic.ai.Node;
import basic.logic.GameLogic;
import basic.logic.Move;
import basic.logic.State;

public class RandomPlayoutPolicy implements DefaultPolicy {

	private GameLogic gl;
	
	private Random rand = new Random();
	
	public RandomPlayoutPolicy(GameLogic gameLogic) {
		this.gl = gameLogic;
	}
	
	@Override
	public int[] playout(Node node) {
		basic.logic.State nextNode = node.getState();
		while (!gl.isTerminal(nextNode)) {
			// all legal moves
			List<Move> moves = gl.legal(nextNode);
			// randomly choose one move
			Move rMove = moves.get(Math.abs(rand.nextInt()) % moves.size());
			// calculate next state
			nextNode = gl.nextState(nextNode, rMove);
		}
		return gl.goal(nextNode);
	}

}
