package basic.logic;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import basic.prepro.STTranslator;

import com.parctechnologies.eclipse.Atom;
import com.parctechnologies.eclipse.CompoundTerm;
import com.parctechnologies.eclipse.CompoundTermImpl;
import com.parctechnologies.eclipse.EclipseEngineOptions;
import com.parctechnologies.eclipse.EclipseException;
import com.parctechnologies.eclipse.OutOfProcessEclipse;

/**
 * GGP Basicplayer of the University of Potsdam
 *
 * Original part of Centurio:
 * Copyright (C) 2008 Felix Maximilian Möller, Marius Schneider and Martin Wegner
 *
 * This file is part of Centurio.
 *
 * Centurio is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * Centurio is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with Centurio. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * This class defines the methods for the game logic, which will be used from the
 * artificial intelligence.
 *
 * @author Felix Maximilian Möller
 * @author Marius Schneider
 * @author Martin Wegner
 * @version 1.1
 */
public class PrologLogic implements GameLogic {

	private OutOfProcessEclipse Eclipse = null;
	private int numberOfPlayers = 0; // number of players in the current game
	private LinkedList<Atom> PlayerList; // list of all players
	private boolean isParallelGame;
	int j;
	
	/**
	 * Constructor <br>
	 * initialize a prolog engine
	 * @param matchID
	 */
	public PrologLogic(String matchID) {
		 //initialize ECLiPSe engine and load game description
		 try {
			EclipseEngineOptions options = new EclipseEngineOptions(new File(""));
			options.setLocalSize(256);
			Eclipse = new OutOfProcessEclipse(options);
			//Setup the randomseed
			int random;
			Random rand = new Random();
			random = rand.nextInt();
			Eclipse.rpc(new CompoundTermImpl("seed", random));
			//load the GameLogic and game description file
			Eclipse.compile(new File("./pl/GameLogic.pl"));
			Eclipse.compile(new File("./pl/" + matchID + ".pl"));
			//disable warnings in Prolog
			Eclipse.rpc("set_stream(warning_output, null)");
			getRoles();
			
		} catch (IOException e) {
			System.err.println("(GameLogic - GameLogic) error message: create eclispe instance or set the seed for random - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
		} catch (EclipseException e) {
			System.err.println("(GameLogic - GameLogic) error message: create eclispe instance or set the seed for random - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
		}
	}
	
	/**
	 * calculates all player of the current game
	 * updates the variable numberOfPlayers
	 * @return all player of the current game in KIF Format,<br>
	 * e.g. :"[com.parctechnologies.eclipse.Atom with [functor=dropper], com.parctechnologies.eclipse.Atom with [functor=player]]" <br><br>
	 * to get the Player in String Format cast each element to CompoundTerm and access with "Object.funktor()"
	 */
	public List<Role> getRoles() {
		if (PlayerList == null) {
			CompoundTerm Query = null;
			try {
				//%% roles(+Roles)bestimmt eine Liste von allen Rollen.
				//gibt zur�ck: "[com.parctechnologies.eclipse.Atom with [functor=dropper], com.parctechnologies.eclipse.Atom with [functor=player]]"
				Query =	Eclipse.rpc("roles(Roles)");
			} catch (IOException e) {
				System.err.println("(GameLogic - getAllPlayers) error message: get a list of all players - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			} catch (EclipseException e) {
				System.err.println("(GameLogic - getAllPlayers) error message: get a list of all players - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			}
			PlayerList = (LinkedList<Atom>) Query.arg(1);
			numberOfPlayers=PlayerList.size();
		}
		List<Role> result = new LinkedList<Role>();
		for (Atom at : PlayerList) {
			result.add(new Role(at));
		}
		return result;
	}
	
	/**
	 * determines in which place the Centurios move stands in the list of moves of all Players
	 * @return all player of the current game in KIF Format,<br>
	 * e.g. :"[com.parctechnologies.eclipse.Atom with [functor=dropper], com.parctechnologies.eclipse.Atom with [functor=player]]" <br><br>
	 * to get the Player in String Format cast each element to CompoundTerm and access with "Object.funktor()"
	 */
	public int findIndex(Role role)	{
		String playerRole = (String)role.get();
		for(Atom Role:PlayerList){
			if(playerRole.equals(Role.functor()))
				return PlayerList.indexOf(Role);
		}
		return -1;
		
	}
	
	/**
	 * approximates if the game is parallel one
	 * @return is true if the game is parallel and false if not
	 */
	public boolean isParallel()	{
		try {
			//%% isParallelGame()
			//%% Prolog predicate checks if more than 1 Player in Initstate have more than one legal move
			Eclipse.rpc("isParallelGame");
		} catch (IOException e) {
			System.err.println("(GameLogic - isParallel) error message: check if the game is a parallel one - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
		} catch (EclipseException e) {
			isParallelGame = false;
			return isParallelGame;
		}
		isParallelGame = true;
		return isParallelGame;
	}

	/**
	 * Plays a whole random Game (sequence from
	 * actions and calculated states) till a
	 * terminal state. 
	 * LinkedList of CoumpoundTerm needs type LinkedList<CompoundTerm>
	 * @param State state where the random game begins
	 * @return [0] reached terminal state(CompoundTerm),<br> 
	 *         [1] used actions in the sequence(LinkedList of CoumpoundTerm)
	 */
	public Object[] randomGame(State current) {
		Object oState = current.get();
		CompoundTerm Query = null;
		Object[] Result = new Object[2];
		//%% randomGame(+CurrentState,-TerminalState,[],-ActionPath)
		try {
			Query =	Eclipse.rpc(new CompoundTermImpl("randomGame", oState, null, new Atom("[]"), null));
		} catch (IOException e) {
			System.err.println("(GameLogic - randomGame(Object State)) error message: play a random game - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(oState));
		} catch (EclipseException e) {
			System.err.println("(GameLogic - randomGame(Object State)) error message: play a random game - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(oState));
		}
		Result[0]=Query.arg(2);
		Result[1]=Query.arg(4);
		return Result;
	}
	
	/**
	 * Plays  n random Games (sequence from actions and calculated states) till a
	 * terminal state. 
	 * @param 	State state where the random game begins
	 * @param	n number of games to play
	 * @return 	Object[i][0] terminal state reached(CompoundTerm),<br> 
	 *			Object[i][1] used actions in the sequence(LinkedList of CoumpoundTerm)<br> 
	 *         	where i stands for the i'th randomgame
	 */
	public Object randomGame(State current, int n)
	{
		Object oState = current.get();
		CompoundTerm Query = null;
		//%% randomGameN(+CurrentState,+N,-ListOfTerminalstatesNActionpaths)
		try {
			Query =	Eclipse.rpc(new CompoundTermImpl("randomGameN", oState, n, null));
		} catch (IOException e) {
			System.err.println("(GameLogic - randomGane(Object State, int n)) error message: play a random game with number" + n + " - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
		} catch (EclipseException e) {
			System.err.println("(GameLogic - randomGame(Object State, int n)) error message: play a random game with number" + n + " - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
		}
	
		return Query.arg(3);
	}
	 
	/**
	* calculates the scores of all players of 
	* a given state
	* @param State given State
	* @return scores of all players
	*/
	public int[] goal(State current)
	{
		Object state = current.get();
		CompoundTerm Query = null;
		int[] Result = null;
		//%% winningStateScoreWithRole(+CurrentState,-Score)
		//ermittelt nur die Punkte der Spieler
		try {
			Query =	Eclipse.rpc(new CompoundTermImpl("scoresOfAllPlayers", state, null));
		} catch (IOException e) {
			System.err.println("(GameLogic - scoresState) error message: calculate the scores of all players - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(state));
		} catch (EclipseException e) {
			System.err.println("(GameLogic - scoresState) error message: calculate the scores of all players - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(state));
		}
				
		Collection<CompoundTerm> ScoreList = new LinkedList<CompoundTerm>();
		ScoreList = (LinkedList<CompoundTerm>) Query.arg(2);
		Result = new int[PlayerList.size()];
		int i = 0;
		
		for (CompoundTerm RoleScorePair : ScoreList){
			
			i = findIndex(new Role(((Atom)RoleScorePair.arg(1)).functor().toString()));
			Result[i] = (Integer) RoleScorePair.arg(2);			
			
		}
		return Result;
	}
	
	/**
	 * calculates all legal moves of a given state
	 * @param State given state
	 * @return all legal moves
	 */
	public List<Move> legal(State current)
	{
		Object state = current.get();
		CompoundTerm Query = null;
		//legalMovesOfAllPlayers(+State,-LegalMovesOfAllPlayers)
		try {
			Query =	Eclipse.rpc(new CompoundTermImpl("legalMovesOfAllPlayersCombination", state, null));
		} catch (IOException e) {
			System.err.println("(GameLogic - getAllLegalMoves) error message: get all legal moves for all players - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(state));
		} catch (EclipseException e) {
			System.err.println("(GameLogic - getAllLegalMoves) error message: get all legal moves for all players - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(state));
		}
		LinkedList<Object> moveList = (LinkedList<Object>) Query.arg(2);
		List<Move> result = new LinkedList<Move>();
		j = 0;
		for (Object obj : moveList) {
//			System.out.println("");
//			System.out.println(j);			
//			System.out.println("1337Move: "+new Move(obj).get().toString());
			result.add(new Move(obj));
//			System.out.println("");
			j++;
		}
		j = 0;
		return result;

	}
	
	/**
	 * calculates the initial state of the game
	 * @return initial state as an Object
	 */
	public State initState()
	{
		CompoundTerm Query = null;
		try {
			//%% init_State()
			Query =	Eclipse.rpc("initState(InitState)");
		} catch (IOException e) {
			System.err.println("(GameLogic - getNextState) error message: compute the next state - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
		} catch (EclipseException e) {
			System.err.println("(GameLogic - getNextState) error message: compute the next state - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
		}
			
		return new State(Query.arg(1));
	}
	
	/**
	 * calculates the next state of a given state and action. <br>
	 * the representation of the state has to be unique!!!
	 * @param State given state
	 * @param Move action, which will be done from the state
	 * @return next state as a String[]
	 */
	public State nextState(State state, Move moves)
	{
		Object oState = state.get();
		Object oMove = moves.get();
		CompoundTerm Query = null;
		try {
			//%% nextState(+Current_State,+Move,-Next_State)
			Query =	Eclipse.rpc(new CompoundTermImpl("nextState", oState, oMove, null));
		} catch (IOException e) {
			System.err.println("(GameLogic - getNextState) error message: compute the next state - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(oState));
			System.err.println("Error in State:\n" + STTranslator.printKIF(oMove));
		} catch (EclipseException e) {
			System.err.println("(GameLogic - getNextState) error message: compute the next state - EclipseException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(oState));
			System.err.println("Error in State:\n" + STTranslator.printKIF(oMove));
		}
			
		return new State(Query.arg(3));
	}
	
	/**
	 * calculates if a state is terminal or not
	 * @param State given state
	 * @return terminal yes or no
	 */
	public boolean isTerminal(State State)
	{
		Object obj = State.get();
		try {
			//%% booleanTerminal(Current_State,Terminal)
			Eclipse.rpc(new CompoundTermImpl("terminalState", obj));
		} catch (IOException e) {
			System.err.println("(GameLogic - isTerminal) error message: check if the current state is terminal - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n" + STTranslator.printKIF(obj));
		} catch (EclipseException e) {
			return false;
		}
		return true;
	}
	
	/**
	 * returns the number of playes in the actual game
	 * @return the number of playes in the actual game
	 */
	public int getNumberOfPlayers() {
		return numberOfPlayers;
	}
	
	/**
	 * return list of all player in the actual game
	 */
	public List<Role> getPlayerList() {
		LinkedList<Role> result = new LinkedList<Role>();
		for (Atom at : PlayerList) {
			result.add(new Role(at));
		}
		return result;
	}
	
	/** capsuled method of destructor */
	public void destroy() {
		destroyEclipse();
	}
	
	/**
	 * destroys the prolog eclipse engine
	 */
	public void destroyEclipse() {
		try {
			Eclipse.destroy();
		} catch (IOException e) {
			System.err.println("(GameLogic - destroyEclipse) error message: destroy GameLogic eclipse - IOException\n\n" + e.getMessage() + "\n\n" + e.toString());
		}
	}

	/**
	 * returns the prolog engine of this instance
	 * shouldn't be used (!)
	 * @return the prolog engine of this instance
	 */
	public OutOfProcessEclipse getEclipse() {
		return Eclipse;
	}

	
}
