package basic.logic;


import java.io.IOException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;

import jpl.Atom;
import jpl.Query;
import jpl.Term;
import jpl.Variable;
import jpl.Util;


public class neuePrologLogic /* extends PrologLogic  */implements GameLogic{
	
	private int numberOfPlayers = 0; // number of players in the current game
	private LinkedList<String> PlayerList; // list of all players
	private boolean isParallelGame;
	String file;
	
	/**
	 * Constructor <br>
	 * initialize a prolog engine
	 * 
	 * @param matchID
	 */
	public neuePrologLogic(String matchID) {
			// load the GameLogic and game description file
		jpl.Query.hasSolution("consult('./pl/swi_GameLogic2.pl')");
		file = "'./pl/"+ matchID +".pl'";
//		System.out.println("\\+");
		String test = "sed -e 's/\\~/backslash\\+/g' "+file+" > bla2.pl";
		Process p;
		try {
			p = new ProcessBuilder("sh","-c",test).start();
			try {
				p.waitFor();
			} catch (InterruptedException e) {
				//  Auto-generated catch block
				e.printStackTrace();
			}
		} catch (IOException e) {
			//  Auto-generated catch block
			e.printStackTrace();
		}
		
		test = "sed -e 's/backslash/\\\\/g' bla2.pl > bla.pl";
		Process p2;
		try {
			p2 = new ProcessBuilder("sh","-c",test).start();
			try {
				p2.waitFor();
			} catch (InterruptedException e) {
				//  Auto-generated catch block
				e.printStackTrace();
			}
		} catch (IOException e) {
			//  Auto-generated catch block
			e.printStackTrace();
		}
		
		jpl.Query.hasSolution("consult('bla.pl')");
		// disable warnings in Prolog
		getRoles();
	}

	/**
	 * calculates all player of the current game updates the variable
	 * numberOfPlayers
	 * 
	 * @return all player of the current game in KIF Format,<br>
	 *         e.g. :
	 *         "[com.parctechnologies.eclipse.Atom with [functor=dropper], com.parctechnologies.eclipse.Atom with [functor=player]]"
	 * <br>
	 * <br>
	 *         to get the Player in String Format cast each element to
	 *         CompoundTerm and access with "Object.funktor()"
	 */
	public List<Role> getRoles() {
		if (PlayerList == null) {
			PlayerList = new LinkedList<String>();
			jpl.Query query = new jpl.Query("getRoles", new jpl.Variable("Roles"));
			@SuppressWarnings("rawtypes")
			Hashtable solution = query.oneSolution();
			String role;
			while ( query.hasMoreSolutions() ){
			    solution = query.nextSolution();
			    role = solution.get("Roles").toString();  //Bsp: score(xplayer, 0)
			    PlayerList.add(role);
			}
			numberOfPlayers = PlayerList.size();
		}
		List<Role> result = new LinkedList<Role>();
		for (String str : PlayerList) {
			result.add(new Role(str));
		}
		return result;
	}

	/**
	 * determines in which place the Centurios move stands in the list of moves
	 * of all Players
	 * 
	 * @return all player of the current game in KIF Format,<br>
	 *         e.g. :
	 *         "[com.parctechnologies.eclipse.Atom with [functor=dropper], com.parctechnologies.eclipse.Atom with [functor=player]]"
	 * <br>
	 * <br>
	 *         to get the Player in String Format cast each element to
	 *         CompoundTerm and access with "Object.funktor()"
	 */
	public int findIndex(Role role) {
//		System.out.println("1337 "+(String) role.get());
		String playerRole = (String) role.get();
		for (String Role : PlayerList) {
//			System.out.println(Role);
			if (playerRole.equals(Role))
				return PlayerList.indexOf(Role);
		}
		return -1;
	}

	/**
	 * approximates if the game is parallel one
	 * 
	 * @return is true if the game is parallel and false if not
	 */
	public boolean isParallel() {
		// %% isParallelGame()
		// %% Prolog predicate checks if more than 1 Player in Initstate
		// have more than one legal move
		jpl.Query q = new jpl.Query("isParallelGame");
		isParallelGame = q.hasSolution();
		return isParallelGame;
	}

	/**
	 * Plays a whole random Game (sequence from actions and calculated states)
	 * till a terminal state. LinkedList of CoumpoundTerm needs type
	 * LinkedList<CompoundTerm>
	 * 
	 * @param State
	 *            state where the random game begins
	 * @return [0] reached terminal state(CompoundTerm),<br>
	 *         [1] used actions in the sequence(LinkedList of CoumpoundTerm)
	 */
/**	public Object[] randomGame(State current) {
		Object oState = current.get();
		CompoundTerm Query = null;
		Object[] Result = new Object[2];
		// %% randomGame(+CurrentState,-TerminalState,[],-ActionPath)
		try {
			Query = Eclipse.rpc(new CompoundTermImpl("randomGame", oState,
					null, new Atom("[]"), null));
		} catch (IOException e) {
			System.err
					.println("(GameLogic - randomGame(Object State)) error message: play a random game - IOException\n\n"
							+ e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n"
					+ STTranslator.printKIF(oState));
		} catch (EclipseException e) {
			System.err
					.println("(GameLogic - randomGame(Object State)) error message: play a random game - EclipseException\n\n"
							+ e.getMessage() + "\n\n" + e.toString());
			System.err.println("Error in State:\n"
					+ STTranslator.printKIF(oState));
		}
		Result[0] = Query.arg(2);
		Result[1] = Query.arg(4);
		return Result;
	}
**/

	/**
	 * Plays n random Games (sequence from actions and calculated states) till a
	 * terminal state.
	 * 
	 * @param State
	 *            state where the random game begins
	 * @param n
	 *            number of games to play
	 * @return Object[i][0] terminal state reached(CompoundTerm),<br>
	 *         Object[i][1] used actions in the sequence(LinkedList of
	 *         CoumpoundTerm)<br>
	 *         where i stands for the i'th randomgame
	 */
/**	public Object randomGame(State current, int n) {
		Object oState = current.get();
		CompoundTerm Query = null;
		// %% randomGameN(+CurrentState,+N,-ListOfTerminalstatesNActionpaths)
		try {
			Query = Eclipse.rpc(new CompoundTermImpl("randomGameN", oState, n,
					null));
		} catch (IOException e) {
			System.err
					.println("(GameLogic - randomGane(Object State, int n)) error message: play a random game with number"
							+ n
							+ " - IOException\n\n"
							+ e.getMessage()
							+ "\n\n" + e.toString());
		} catch (EclipseException e) {
			System.err
					.println("(GameLogic - randomGame(Object State, int n)) error message: play a random game with number"
							+ n
							+ " - EclipseException\n\n"
							+ e.getMessage()
							+ "\n\n" + e.toString());
		}

		return Query.arg(3);
	}
**/
	
	/**
	 * calculates the scores of all players of a given state
	 * 
	 * @param State
	 *            given State
	 * @return scores of all players
	 */
	public int[] goal(State current) {
		jpl.Term state = (jpl.Term) current.get();
		jpl.Variable score = new jpl.Variable("Score");
		int[] Result = null;
		int index;
		String scoreStr;
		// %% scoresOfAllPlayers(+CurrentState,-Score)
		// ermittelt nur die Punkte der Spieler
		jpl.Query q = new jpl.Query(/*"scoresOfAllPlayers"*/"bla",new jpl.Term[] {state, score} );
		@SuppressWarnings("rawtypes")
		Hashtable solution = q.oneSolution();
		Result = new int[PlayerList.size()];
		while ( q.hasMoreSolutions() ){
		    solution = q.nextSolution();
		    scoreStr = solution.get("Score").toString();  //Bsp: score(xplayer, 0)
//		    System.out.println("scoreStr "+scoreStr);
		    //Für alle Rollen...
		    for(String player : PlayerList){
		    	if(scoreStr.matches(".*"+player+".*")){						// Überprüfen um welche Rolle es sich handelt
		    		index = findIndex(new Role(player));							//Index der Role finden
		    		Result[index] = Integer.parseInt(scoreStr.replaceAll("\\D", "")); //Punkte zuweisen
//		    		System.out.println(index+ "\t"+ Result[index]);
		    	}
		    }
		}
//		for( int i = 0; i < PlayerList.size(); i++ ){
//			System.out.println("Role: "+PlayerList.get(i)+"	Score: "+Result[i]+"	i: "+i);
//		}
		return Result;
	}

	/**
	 * calculates all legal moves of a given state
	 * 
	 * @param State
	 *            given state
	 * @return all legal moves
	 */
	public List<Move> legal(State current) {
		List<Move> result = new LinkedList<Move>();
		jpl.Term term;
		jpl.Term state = (jpl.Term) current.get();
		jpl.Variable legalMoves = new jpl.Variable("LegalMovesOfAllPlayers");
		jpl.Term arg[] = {state, legalMoves};
		// legalMovesOfAllPlayers(+State,-LegalMovesOfAllPlayers)
		// LegalMovesOfAllPlayers gibt einen Move(Liste) pro Solution
		jpl.Query q = new jpl.Query("legalMovesOfAllPlayersCombination", arg/*new Term[] {state, legalMoves}*/);
		@SuppressWarnings("rawtypes")
		Hashtable solution; //= q.oneSolution();
		while ( q.hasMoreSolutions() ){
		    solution = q.nextSolution();
		    term = (jpl.Term) solution.get("LegalMovesOfAllPlayers");
//		    System.out.println( "LegalMove = " + term );
		    result.add(new Move(term));  //Term zur Liste hinzufügen z.B: '.'(mark(1, 1), '.'(noop, []))
		}		
		return result;
	}

	/**
	 * calculates the initial state of the game
	 * 
	 * @return initial state as an Object
	 */
	public State initState() {
		 jpl.Query query = new jpl.Query("initState", new jpl.Variable("InitState"));
		 @SuppressWarnings("rawtypes")
		 Hashtable solution = query.oneSolution();
		 jpl.Term initState = (jpl.Term) solution.get("InitState");
		 return new State(initState);
	}

	/**
	 * calculates the next state of a given state and action. <br>
	 * the representation of the state has to be unique!!!
	 * 
	 * @param State
	 *            given state
	 * @param Move
	 *            action, which will be done from the state
	 * @return next state as a String[]
	 */
	public State nextState(State state, Move moves) {
//		System.out.println("Moves: " + moves.get().toString());  // (MARK 1 1) NOOP) [mark(2,3), noop]
		jpl.Term oState = (jpl.Term) state.get();
		jpl.Term oMove =  (jpl.Term) moves.get();
		jpl.Variable State = new jpl.Variable("State");
		// %% nextState(+Current_State,+Move,-Next_State)
		jpl.Query q = new jpl.Query("nextState", new jpl.Term[]{oState, oMove, State}); 
		@SuppressWarnings("rawtypes")
		Hashtable solution = q.oneSolution();
		jpl.Term nextState = (jpl.Term) solution.get("State");
//		System.out.println("NextState: "+nextState.toString());
		return new State(nextState);
	}

	/**
	 * calculates if a state is terminal or not
	 * 
	 * @param State
	 *            given state
	 * @return terminal yes or no
	 */
	public boolean isTerminal(State State) {
		jpl.Term t = (jpl.Term) State.get();
		// %% booleanTerminal(Current_State,Terminal)
		jpl.Query query = new jpl.Query("terminalState", t); 
		boolean b = query.hasSolution();
//		System.out.println("terminal: "+b);
		return b;
	}

	/**
	 * returns the number of playes in the actual game
	 * 
	 * @return the number of playes in the actual game
	 */
	public int getNumberOfPlayers() {
		return numberOfPlayers;
	}

	/**
	 * return list of all player in the actual game
	 */
	public List<Role> getPlayerList() {
		LinkedList<Role> result = new LinkedList<Role>();
		for (String s : PlayerList) {
			result.add(new Role(s));
		}
		return result;
	}

	/** capsuled method of destructor */
	public void destroy() {
		jpl.JPL.halt();
	}
}